@@ -1,14 +1,14 @@
 package Imager::Preprocess;
+use 5.006;
 use strict;
-require Exporter;
-use vars qw(@ISA @EXPORT $VERSION);
+use warnings;
+use Exporter "import";
 use Getopt::Long;
 use Text::ParseWords;
 
-@EXPORT = qw(preprocess);
-@ISA = qw(Exporter);
+our @EXPORT = qw(preprocess);
 
-$VERSION = "1.000";
+our $VERSION = "1.004";
 
 sub preprocess {
   unshift @ARGV, grep /^-/, shellwords($ENV{IMAGER_PREPROCESS_OPTS})
@@ -32,18 +32,20 @@ sub preprocess {
   my $code_line;
   my @out;
   my $failed;
+  my %replace;
 
   push @out, 
     "#define IM_ROUND_8(x) ((int)((x)+0.5))\n",
     "#define IM_ROUND_double(x) (x)\n",
     "#define IM_LIMIT_8(x) ((x) < 0 ? 0 : (x) > 255 ? 255 : (x))\n",
-    "#define IM_LIMIT_double(x) ((x) < 0.0 ? 0.0 : (x) > 1.0 ? 1.0 : (x))\n";
+    "#define IM_LIMIT_double(x) ((x) < 0.0 ? 0.0 : (x) > 1.0 ? 1.0 : (x))\n",
+    "#define IM_LIMIT_16(x) ((x) < 0 ? 0 : (x) > 65535 ? 65535 : (x))\n";
   push @out, "#line 1 \"$src\"\n" if $keep_lines;
   while (defined(my $line = <SRC>)) {
     if ($line =~ /^\#code\s+(\S.+)$/) {
       $save_code
 	and do { warn "$src:$code_line:Unclosed #code block\n"; ++$failed; };
-      
+
       $cond = $1;
       $cond_line = $.;
       $code_line = $. + 1;
@@ -52,7 +54,7 @@ sub preprocess {
     elsif ($line =~ /^\#code\s*$/) {
       $save_code
 	and do { warn "$src:$code_line:Unclosed #code block\n"; ++$failed; };
-      
+
       $cond = '';
       $cond_line = 0;
       $code_line = $. + 1;
@@ -61,7 +63,7 @@ sub preprocess {
     elsif ($line =~ /^\#\/code\s*$/) {
       $save_code
 	or do { warn "$src:$.:#/code without #code\n"; ++$failed; next; };
-      
+
       if ($cond) {
 	push @out, "#line $cond_line \"$src\"\n" if $keep_lines;
 	push @out, "  if ($cond) {\n";
@@ -73,6 +75,11 @@ sub preprocess {
         "#define IM_FILL_COMBINE(fill) ((fill)->combine)\n",
 	"#undef IM_FILL_FILLER\n",
         "#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_color)\n";
+      for my $key (sort keys %replace) {
+	push @out,
+	  "#undef $key\n",
+	  "#define $key $replace{$key}[0]\n";
+      }
       push @out, "#line $code_line \"$src\"\n" if $keep_lines;
       push @out, byte_samples(@code);
       push @out, "  }\n", "  else {\n"
@@ -83,6 +90,11 @@ sub preprocess {
         "#define IM_FILL_COMBINE(fill) ((fill)->combinef)\n",
 	"#undef IM_FILL_FILLER\n",
         "#define IM_FILL_FILLER(fill) ((fill)->f_fill_with_fcolor)\n";
+      for my $key (sort keys %replace) {
+	push @out,
+	  "#undef $key\n",
+	  "#define $key $replace{$key}[1]\n";
+      }
       push @out, "#line $code_line \"$src\"\n" if $keep_lines;
       push @out, double_samples(@code);
       push @out, "  }\n"
@@ -94,40 +106,54 @@ sub preprocess {
     elsif ($save_code) {
       push @code, $line;
     }
+    elsif ($line =~ /^\#!\s*define\s+(\w+)\s+(\w+)\s+(\w+)\s*$/) {
+      exists $replace{$1}
+	and die "Duplicate definition of $1\n";
+      $replace{$1} = [ $2, $3 ];
+    }
     else {
       push @out, $line;
     }
   }
-  
+
   if ($save_code) {
     warn "$src:$code_line:#code block not closed by EOF\n";
     ++$failed;
   }
 
   close SRC;
-  
+
   $failed 
     and die "Errors during parsing, aborting\n";
-  
+
   open DEST, "> $dest"
     or die "Cannot open $dest: $!\n";
   print DEST @out;
   close DEST;
 }
-  
+
 sub byte_samples {
   # important we make a copy
   my @lines = @_;
-  
+
   for (@lines) {
     s/\bIM_GPIX\b/i_gpix/g;
     s/\bIM_GLIN\b/i_glin/g;
     s/\bIM_PPIX\b/i_ppix/g;
     s/\bIM_PLIN\b/i_plin/g;
+    s/\bIM_GSLIN\b/i_gslin/g;
+    s/\bIM_PSLIN\b/i_pslin/g;
     s/\bIM_GSAMP\b/i_gsamp/g;
+    s/\bIM_GSAMP_ASSERT\b/i_gsamp_assert/g;
+    s/\bIM_PSAMP\b/i_psamp/g;
+    s/\bIM_GSLIN\b/i_gslin/g;
+    s/\bIM_GSLIN_ASSERT\b/i_gslin_assert/g;
+    s/\bIM_PSLIN\b/i_pslin/g;
     s/\bIM_SAMPLE_MAX\b/255/g;
     s/\bIM_SAMPLE_MAX2\b/65025/g;
     s/\bIM_SAMPLE_T/i_sample_t/g;
+    s/\bIM_LIN_SAMPLE_T\b/i_sample16_t/g;
+    s/\bIM_LIN_SAMPLE_MAX\b/65535/g;
     s/\bIM_COLOR\b/i_color/g;
     s/\bIM_WORK_T\b/int/g;
     s/\bIM_Sf\b/"%d"/g;
@@ -136,26 +162,37 @@ sub byte_samples {
     s/\bIM_ROUND\(/IM_ROUND_8(/g;
     s/\bIM_ADAPT_COLORS\(/i_adapt_colors(/g;
     s/\bIM_LIMIT\(/IM_LIMIT_8(/g;
+    s/\bIM_LIN_LIMIT\(/IM_LIMIT_16(/g;
     s/\bIM_RENDER_LINE\(/i_render_line(/g;
     s/\bIM_FILL_COMBINE_F\b/i_fill_combine_f/g;
+    s/\bIM_ABS\b/abs/g;
   }
-  
+
   @lines;
 }
 
 sub double_samples {
   # important we make a copy
   my @lines = @_;
-  
+
   for (@lines) {
     s/\bIM_GPIX\b/i_gpixf/g;
     s/\bIM_GLIN\b/i_glinf/g;
     s/\bIM_PPIX\b/i_ppixf/g;
     s/\bIM_PLIN\b/i_plinf/g;
+    s/\bIM_GSLIN\b/i_gslinf/g;
+    s/\bIM_PSLIN\b/i_pslinf/g;
     s/\bIM_GSAMP\b/i_gsampf/g;
+    s/\bIM_GSAMP_ASSERT\b/i_gsampf_assert/g;
+    s/\bIM_PSAMP\b/i_psampf/g;
+    s/\bIM_GSLIN\b/i_gslinf/g;
+    s/\bIM_GSLIN_ASSERT\b/i_gslinf_assert/g;
+    s/\bIM_PSLIN\b/i_pslinf/g;
     s/\bIM_SAMPLE_MAX\b/1.0/g;
     s/\bIM_SAMPLE_MAX2\b/1.0/g;
     s/\bIM_SAMPLE_T/i_fsample_t/g;
+    s/\bIM_LIN_SAMPLE_T\b/i_fsample_t/g;
+    s/\bIM_LIN_SAMPLE_MAX\b/1.0/g;
     s/\bIM_COLOR\b/i_fcolor/g;
     s/\bIM_WORK_T\b/double/g;
     s/\bIM_Sf\b/"%f"/g;
@@ -164,8 +201,10 @@ sub double_samples {
     s/\bIM_ROUND\(/IM_ROUND_double(/g;
     s/\bIM_ADAPT_COLORS\(/i_adapt_fcolors(/g;
     s/\bIM_LIMIT\(/IM_LIMIT_double(/g;
+    s/\bIM_LIN_LIMIT\(/IM_LIMIT_double(/g;
     s/\bIM_RENDER_LINE\(/i_render_linef(/g;
     s/\bIM_FILL_COMBINE_F\b/i_fill_combinef_f/g;
+    s/\bIM_ABS\b/fabs/g;
   }
 
   @lines;
@@ -234,6 +273,14 @@ The end of a sample-independent section of code is terminated by:
 
 #/code without a starting #code is an error.
 
+You can also define extra sample-size dependent macros with C<#!define>:
+
+=over
+
+C<#!define> I<common-name> I<eight-bit-name> I<floating-point-name>
+
+=back
+
 The following types and values are defined in a #code section:
 
 =over
@@ -263,6 +310,22 @@ becomes i_gpix() or i_gpixf() as appropriate.
 
 =item *
 
+IM_GSAMP_ASSERT(C<im>, C<l>, C<r>, C<y>, C<samples>, C<chans>, C<chan_count>)
+
+=item *
+
+IM_GSLIN(C<im>, C<l>, C<r>, C<y>, C<samples>, C<chans>, C<chan_count>)
+
+=item *
+
+IM_GSLIN_ASSERT(C<im>, C<l>, C<r>, C<y>, C<samples>, C<chans>, C<chan_count>)
+
+=item *
+
+IM_PSLIN(C<im>, C<l>, C<r>, C<y>, C<samples>, C<chans>, C<chan_count>)
+
+=item *
+
 IM_ADAPT_COLORS(C<dest_channels>, C<src_channels>, C<colors>, C<count>)
 
 Call i_adapt_colors() or i_adapt_fcolors().
@@ -279,7 +342,11 @@ object.
 
 =item *
 
-IM_SAMPLE_MAX - maximum value for a sample
+IM_ABS(sample) - calculate the absolute value of an IM_WORK_T value. (abs(), fabs())
+
+=item *
+
+IM_SAMPLE_MAX - maximum value for a sample (255, 1.0)
 
 =item *
 
@@ -287,23 +354,31 @@ IM_SAMPLE_MAX2 - maximum value for a sample, squared
 
 =item *
 
-IM_SAMPLE_T - type of a sample (i_sample_t or i_fsample_t)
+IM_SAMPLE_T - type of a sample (i_sample_t, i_fsample_t)
 
 =item *
 
-IM_COLOR - color type, either i_color or i_fcolor.
+IM_LIN_SAMPLE_T - type of a linear sample (i_sample16_t, i_fsample_t)
 
 =item *
 
-IM_WORK_T - working sample type, either int or double.
+IM_LIN_SAMPLE_MAX - maximum value for a linear sample (65535, 1.0)
 
 =item *
 
-IM_Sf - format string for the sample type, C<"%d"> or C<"%f">.
+IM_COLOR - color type (i_color, i_fcolor).
 
 =item *
 
-IM_Wf - format string for the work type, C<"%d"> or C<"%f">.
+IM_WORK_T - working sample type, (int, double).
+
+=item *
+
+IM_Sf - format string for the sample type, (C<"%d">, C<"%f">).
+
+=item *
+
+IM_Wf - format string for the work type, (C<"%d">, C<"%f">).
 
 =item *
 
@@ -313,6 +388,12 @@ IM_SUFFIX(identifier) - adds _8 or _double onto the end of identifier.
 
 IM_EIGHT_BIT - this is a macro defined only in 8-bit/sample code.
 
+  #ifdef IM_EIGHT_BIT
+   /* 8-bit per sample only code */
+  #else
+   /* double per sample only code */
+  #endif
+
 =back
 
 Other types, functions and values may be added in the future.
